/* development styles */
window.development-version headerbar {
  background: transparent -gtk-icontheme("system-run-symbolic") 80% 0/128px 128px no-repeat,
              linear-gradient(to left,
                              mix(@theme_fg_color, @theme_bg_color, 0.5) 0%,
                              @theme_bg_color 25%);
  color: alpha(@theme_fg_color, 0.2);
}

window.development-version headerbar box label {
  color: @theme_fg_color;
}

button.sysprofprofilermenubutton.popup.suggested-action:checked {
  background: shade(@theme_selected_bg_color,0.9);
  border: 1px solid shade(@theme_selected_bg_color, 0.8);
  color: alpha(@theme_selected_fg_color,0.9);
}
