popover list row {
  padding: 6px 10px 6px 10px;
  border-bottom: 1px solid alpha(@borders, 0.2);
}

popover list row:last-child {
  border-bottom: none;
}

popover scrolledwindow {
  border-top: 1px solid alpha(@borders, 0.75);
}

visualizers list {
  background: @theme_bg_color;
}

visualizers list row:backdrop,
visualizers list row:last-child {
  border-bottom: none;
}

ticks {
  color: alpha(@theme_fg_color, 0.4);
}
